//
// Copyright (C) 2005 Mirko Nasato <mirko@artofsolving.com>
//
// Distributable under the terms of the LGPL license.
// See http://www.gnu.org/copyleft/lesser.html
//
package net.sf.davcache;

import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.QName;
import org.dom4j.io.OutputFormat;
import org.dom4j.io.SAXReader;
import org.dom4j.io.XMLWriter;

/**
 * XML response to a PROPFIND request.
 * <p>
 * See <a
 * href="http://www.webdav.org/specs/rfc2518.html#METHOD_PROPFIND">rfc2518</a>
 * for the specifications.
 * 
 * @author Mirko Nasato
 */
public class PropFindResponse
{
   public static final int CODE_MULTI_STATUS = 207;

   private static final String STATUS_OK = "HTTP/1.1 200 OK";

   private static final String STATUS_NOT_FOUND = "HTTP/1.1 404 Not Found";

   private Document document;

   private Element multistatus;

   private List/* QName */requestedProperties;

   public PropFindResponse(HttpServletRequest request) throws IOException, DocumentException
   {
      if (request.getContentLength() > 0)
      {
         requestedProperties = parseRequestedProperties(request);
      }
      else
      {
         // rfc2518 8.1 PROPFIND: A client may choose not to submit a request body.
         // An empty PROPFIND request body MUST be treated as a request for the names and values of all properties.
      }
      document = DocumentHelper.createDocument();
      multistatus = document.addElement(DavXmlUtils.toDavQName("multistatus"));
   }

   public void addResource(String path, boolean collection, Map props)
   {
      List missingProperties = new ArrayList();
      if (requestedProperties != null)
      {
         missingProperties.addAll(requestedProperties);
      }
      if (collection && !path.endsWith("/"))
      {
         path += "/";
      }
      Element response = DavXmlUtils.addDavElement(multistatus, "response");
      DavXmlUtils.addDavElement(response, "href", path);
      Element propStatOk = DavXmlUtils.addDavElement(response, "propstat");
      Element propOk = DavXmlUtils.addDavElement(propStatOk, "prop");

      QName qResourceType = DavXmlUtils.toDavQName("resourcetype");
      if (requestedProperties == null || requestedProperties.contains(qResourceType))
      {
         Element resourceType = propOk.addElement(qResourceType);
         if (collection)
         {
            DavXmlUtils.addDavElement(resourceType, "collection");
         }
         missingProperties.remove(qResourceType);
      }

      for (Iterator iter = props.entrySet().iterator(); iter.hasNext();)
      {
         Map.Entry entry = (Map.Entry) iter.next();
         QName qProperty = DavXmlUtils.toDavQName((String) entry.getKey());
         if (missingProperties.contains(qProperty))
         {
            Element property = propOk.addElement(qProperty);
            property.setText(entry.getValue().toString());
            missingProperties.remove(qProperty);
         }
      }
      DavXmlUtils.addDavElement(propStatOk, "status", STATUS_OK);

      if (missingProperties.size() > 0)
      {
         Element propStatNotFound = DavXmlUtils.addDavElement(response, "propstat");
         Element propNotFound = DavXmlUtils.addDavElement(propStatNotFound, "prop");
         for (Iterator iter = missingProperties.iterator(); iter.hasNext();)
         {
            QName qProperty = (QName) iter.next();
            if ("DAV:".equals(qProperty.getNamespaceURI()))
            {
               // use "D:" prefix as we do with all other DAV properties
               qProperty = DavXmlUtils.toDavQName(qProperty.getName());
            }
            propNotFound.addElement(qProperty);
         }
         DavXmlUtils.addDavElement(propStatNotFound, "status", STATUS_NOT_FOUND);
      }
   }

   // debug only
   public void write(OutputStream output) throws IOException
   {
      XMLWriter writer = new XMLWriter(output, new OutputFormat("  ", true));
      writer.write(document);
   }

   public void send(HttpServletResponse response) throws IOException
   {
      response.setStatus(CODE_MULTI_STATUS);
      response.setContentType("text/xml;charset=UTF-8");
      XMLWriter writer = new XMLWriter(response.getOutputStream(), new OutputFormat("    ", true));
      writer.write(document);
   }

   public String asXml()
   {
      return document.asXML();
   }

   private List/* QName */parseRequestedProperties(HttpServletRequest request) throws IOException, DocumentException
   {
      SAXReader reader = new SAXReader();
      Document document = reader.read(request.getInputStream());
      Element propfind = document.getRootElement();
      if (propfind == null)
      {
         throw new DocumentException("no root element");
      }
      if (!propfind.getName().equals("propfind"))
      {
         throw new DocumentException("root element is not 'propfind'");
      }
      List propfindElements = propfind.elements();
      if (propfindElements.size() != 1)
      {
         throw new DocumentException("'propfind' contains " + propfindElements.size() + " nodes instead of 1");
      }
      Element prop = (Element) propfindElements.get(0);
      if (prop.getName().equals("allprop"))
      {
         // request for all properties; see http://www.webdav.org/specs/rfc2518.html#rfc.section.8.1.2
         return null;
      }
      if (!prop.getName().equals("prop"))
      {
         throw new DocumentException("'propfind' child element is neither 'prop' nor 'allprop'");
      }
      List properties = new ArrayList();
      for (Iterator iterator = prop.elementIterator(); iterator.hasNext();)
      {
         Element property = (Element) iterator.next();
         properties.add(property.getQName());
      }
      return properties;
   }
}
