//
// Copyright (C) 2005 Mirko Nasato <mirko@artofsolving.com>
//
// Distributable under the terms of the LGPL license.
// See http://www.gnu.org/copyleft/lesser.html
//
package net.sf.davcache;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.jboss.cache.CacheException;
import org.jboss.cache.TreeCacheMBean;

/**
 * FileSystem-like wrapper for TreeCache.
 * <p>
 * Any given path can denote either a 'directory' ('collection' in the Webdav spec)
 * or a regular 'file' ('non collection resource' in the Webdav spec).
 * 
 * @author Mirko Nasato
 */
public class TreeCacheStore implements TreeCacheStoreMBean
{

   private static class FqnAndKey
   {
      private String fqn;

      private String key;

      public FqnAndKey(String fqn, String key)
      {
         this.fqn = fqn;
         this.key = key;
      }

      public String getFqn()
      {
         return fqn;
      }

      public String getKey()
      {
         return key;
      }
   }

   //private static Logger logger = Logger.getLogger(TreeCacheStore.class);
   
   private static final String CONTENT_PATH = "__WEBDAV_CONTENT__";
   private static final String PROPERTIES_PATH = "__WEBDAV_PROPERTIES__";   

   private static final DateFormat httpDateFormat = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss z");

   private FileTypes fileTypes = new FileTypes();

   private String treeCacheJndiName;
   private TreeCacheMBean treeCache;

   public TreeCacheStore() { }

   /**
    * @see net.sf.davcache.TreeCacheStoreMBean#setTreeCache(org.jboss.cache.TreeCacheMBean)
    */
   public void setTreeCache(TreeCacheMBean treeCache) throws CacheException
   {
      this.treeCache = treeCache;
   }

   /**
    * @see net.sf.davcache.TreeCacheStoreMBean#setTreeCacheJndiName(java.lang.String)
    */
   public void setTreeCacheJndiName(String treeCacheJndiName)
   {
      this.treeCacheJndiName = treeCacheJndiName;
   }

   public void start() throws NamingException, CacheException
   {
      if (treeCache == null && treeCacheJndiName == null)
         throw new IllegalStateException("no TreeCache instance nor TreeCacheJndiName set");

      if (treeCache == null)
      {
         Context jndi = new InitialContext();
         treeCache = (TreeCacheMBean) jndi.lookup(treeCacheJndiName);
      }
      treeCache.put(CONTENT_PATH, null);
      treeCache.put(PROPERTIES_PATH, null);
   }

   public void stop()
   {
      // nothing to do
   }

   private String getContentPath(String path)
   {
      return CONTENT_PATH + "/" + path;
   }

   private String getPropertiesPath(String path)
   {
      return PROPERTIES_PATH + "/" + path;
   }

   private FqnAndKey splitFqnAndKey(String path)
   {
      String fqn = "";
      String key = path;
      int slash = path.lastIndexOf('/');
      if (slash != -1)
      {
         fqn = path.substring(0, slash);
         key = path.substring(slash + 1);
      }
      return new FqnAndKey(fqn, key);
   }

   public boolean isDirectory(String path)
   {
      return treeCache.exists(getContentPath(path));
   }

   public boolean isFile(String path)
   {
      FqnAndKey pair = splitFqnAndKey(getContentPath(path));
      return treeCache.exists(pair.getFqn(), pair.getKey());
   }

   public boolean exists(String path)
   {
      return isDirectory(path) || isFile(path);
   }

   public Map getProperties(String path) throws CacheException
   {
      Map properties = new HashMap();
      String propertiesPath = getPropertiesPath(path);
      Set names = treeCache.getKeys(propertiesPath);
      if (names != null)
      {
         for (Iterator iter = names.iterator(); iter.hasNext();)
         {
            String name = (String) iter.next();
            Object value = treeCache.get(propertiesPath, name);
            properties.put(name, value);
         }
      }
      return properties;
   }

   public String getContentType(String path) throws CacheException
   {
      return (String) getProperties(path).get("getcontenttype");
   }

   public Set getFileNames(String path) throws CacheException
   {
      Set names = treeCache.getKeys(getContentPath(path));
      if (names == null)
      {
         names = new HashSet();  // never return null
      }
      return names;
   }

   public Set getDirectoryNames(String path) throws CacheException
   {
      Set names = treeCache.getChildrenNames(getContentPath(path));
      if (names != null)
      {
         names = new HashSet(names);
         names.remove(PROPERTIES_PATH);
      } else {
         names = new HashSet();  // never return null
      }
      return names;
   }

   public boolean isDirectoryEmpty(String path) throws CacheException
   {
      String contentPath = getContentPath(path);
      Set leaves = getFileNames(contentPath);
      Set branches = getDirectoryNames(contentPath);
      return (leaves == null || leaves.isEmpty()) && (branches == null || branches.isEmpty());
   }

   public void createDirectory(String path) throws CacheException
   {
      treeCache.put(getContentPath(path), null);
   }

   public void createFile(String path, byte[] binaryContent) throws CacheException
   {
      Map properties = new HashMap();
      properties.put("getcontentlength", Long.toString(binaryContent.length));
      properties.put("getcontenttype", fileTypes.getContentType(path));
      properties.put("getlastmodified", httpDateFormat.format(new Date()));
      FqnAndKey pair = splitFqnAndKey(getContentPath(path));
      treeCache.put(pair.getFqn(), pair.getKey(), binaryContent);
      treeCache.put(getPropertiesPath(path), properties);
   }

   public byte[] getFileContent(String path) throws CacheException
   {
      FqnAndKey pair = splitFqnAndKey(getContentPath(path));
      return (byte[]) treeCache.get(pair.getFqn(), pair.getKey());
   }

   public String getFileContentAsString(String path) throws CacheException
   {
      if (!isFile(path))
         return null;
      return new String(getFileContent(path));
   }

   public void deleteDirectory(String path) throws CacheException
   {
      treeCache.remove(getContentPath(path));
      String propertiesPath = getPropertiesPath(path);
      if (treeCache.exists(propertiesPath))
      {
         treeCache.remove(propertiesPath);
      }
      if (path.equals(""))
      {
         // recreate root paths 
         treeCache.put(CONTENT_PATH, null);
         treeCache.put(PROPERTIES_PATH, null);
      }
   }

   public void deleteFile(String path) throws CacheException
   {
      FqnAndKey pair = splitFqnAndKey(getContentPath(path));
      treeCache.remove(pair.getFqn(), pair.getKey());
      treeCache.remove(getPropertiesPath(path));
   }

   public void copyDirectory(String sourcePath, String destinationPath) throws CacheException
   {
      createDirectory(destinationPath);
      Set subDirNames = getDirectoryNames(sourcePath);
      for (Iterator it = subDirNames.iterator(); it.hasNext();) {
         String dirName = (String) it.next();
         copyDirectory(sourcePath + "/" + dirName, destinationPath + "/" + dirName);
      }
      Set fileNames = getFileNames(sourcePath);
      for (Iterator it = fileNames.iterator(); it.hasNext();) {
         String fileName = (String) it.next();
         copyFile(sourcePath + "/" + fileName, destinationPath + "/" + fileName);
      }
   }

   public void copyFile(String path, String destinationPath) throws CacheException
   {
      byte[] fileContent = getFileContent(path);
      createFile(destinationPath, fileContent);
   }

   public void moveDirectory(String path, String destinationPath) throws CacheException
   {
      copyDirectory(path, destinationPath);
      deleteDirectory(path);
   }

   public void moveFile(String path, String destinationPath) throws CacheException
   {
      copyFile(path, destinationPath);
      deleteFile(path);
   }

}
