//
// Copyright (C) 2005 Mirko Nasato <mirko@artofsolving.com>
//
// Distributable under the terms of the LGPL license.
// See http://www.gnu.org/copyleft/lesser.html
//
package net.sf.davcache;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;
import java.util.Set;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.dom4j.DocumentException;
import org.jboss.cache.CacheException;
import org.jboss.logging.Logger;

/**
 * A WebDAV interface to JBossCache.
 * 
 * @author Mirko Nasato
 */
public class WebdavServlet extends HttpServlet
{
   private static final int SC_INSUFFICIENT_STORAGE = 507;
   
   private static final int DEPTH_INFINITY = -1;

   private static final Logger logger = Logger.getLogger(WebdavServlet.class);

   private String treeCacheStoreJndiName;
   private TreeCacheStoreMBean store;
   private int maxFileSize = -1;
   
   public void init(ServletConfig servletConfig) throws ServletException
   {
      super.init(servletConfig);
      
      String maxFileSizeString = servletConfig.getInitParameter("MaxFileSize");
      if (maxFileSizeString != null)
      {
         maxFileSize = Integer.parseInt(maxFileSizeString);
         logger.info("maxFileSize set to " + maxFileSize + " bytes");
      }
      
      treeCacheStoreJndiName = servletConfig.getInitParameter("JndiName"); 
   }

   private void initStore() throws ServletException
   {
      if (treeCacheStoreJndiName != null)
      {
         try
         {
            Context jndi = new InitialContext();            
            store = (TreeCacheStoreMBean) jndi.lookup(treeCacheStoreJndiName);
         }
         catch (NamingException e)
         {
            throw new ServletException("failed to retrieve TreeCache from JNDI name '" + treeCacheStoreJndiName + "'");
         }
      }
      else
      {
         store = (TreeCacheStoreMBean) getServletContext().getAttribute(TreeCacheStore.class.getName());

         if (store == null)
            throw new ServletException("TreeCacheStore not found: neither JndiName set nor attribute '" + TreeCacheStore.class.getName() + "' found in servlet context scope");
      }
   }

   protected void service(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException
   {
      if (store == null)
         initStore();

      String method = request.getMethod().toUpperCase();
      if (method.equals("COPY"))
      {
         doCopy(request, response);
      }
      else if (method.equals("MKCOL"))
      {
         doMkCol(request, response);
      }
      else if (method.equals("MOVE"))
      {
         doMove(request, response);
      }
      else if (method.equals("PROPFIND"))
      {
         doPropFind(request, response);
      }
      else if (method.equals("PROPPATCH"))
      {
         doPropPatch(request, response);
      }
      else
      {
         super.service(request, response);
      }
   }

   protected void doCopy(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException
   {
      String sourcePath = getRelativePath(request);
      if (!store.exists(sourcePath))
      {
         response.sendError(HttpServletResponse.SC_NOT_FOUND);
         return;
      }
      String absoluteDestinationPath = request.getHeader("Destination");
      String destinationPath = relativizePath(request, absoluteDestinationPath);
      if (destinationPath.equals(sourcePath))
      {
         response.sendError(HttpServletResponse.SC_FORBIDDEN, "Source and destination are the same");
         return;
      }
      String parentPath = getParentPath(destinationPath);
      if (parentPath != null && !store.exists(parentPath))
      {
         logger.debug("cannot copy to destination; parent does not exist: '" + destinationPath + "'");
         response.sendError(HttpServletResponse.SC_CONFLICT, "Conflict: destination parent path does not exist");
         return;
      }
      boolean overwrite = "T".equalsIgnoreCase(request.getHeader("Overwrite"));
      boolean destinationAlreadyExists = store.exists(destinationPath);
      if (!overwrite && destinationAlreadyExists)
      {
         response.sendError(HttpServletResponse.SC_PRECONDITION_FAILED, "Resource exists and will not be overwritten: " + destinationPath);
         return;
      }
      try
      {
         if (store.isFile(sourcePath))
         {
            logger.debug("copying file '" + sourcePath + "' to '" + destinationPath + "'");
            store.copyFile(sourcePath, destinationPath);
            if (destinationAlreadyExists)
            {
               response.setStatus(HttpServletResponse.SC_NO_CONTENT);
            }
            else
            {
               response.setStatus(HttpServletResponse.SC_CREATED);
            }
         }
         else if (store.isDirectory(sourcePath))
         {
            logger.debug("copying directory '" + sourcePath + "' to '" + destinationPath + "'");
            store.copyDirectory(sourcePath, destinationPath);
            if (destinationAlreadyExists)
               response.setStatus(HttpServletResponse.SC_NO_CONTENT);
            else
               response.setStatus(HttpServletResponse.SC_CREATED);
         }
      }
      catch (CacheException e)
      {
         throw new ServletException(e);
      }
   }

   protected void doMove(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException
   {
      String path = getRelativePath(request);
      if (!store.exists(path))
      {
         response.sendError(HttpServletResponse.SC_NOT_FOUND);
         return;
      }
      String absoluteDestinationPath = request.getHeader("Destination");
      String destinationPath = relativizePath(request, absoluteDestinationPath);
      boolean overwrite = "T".equalsIgnoreCase(request.getHeader("Overwrite"));
      boolean destinationAlreadyExists = store.exists(destinationPath);
      if (!overwrite && destinationAlreadyExists)
      {
         response.sendError(HttpServletResponse.SC_PRECONDITION_FAILED, "Resource exists and will not be overwritten: " + destinationPath);
         return;
      }
      try
      {
         if (store.isFile(path))
         {
            logger.debug("moving file from '" + path + "' to '" + destinationPath + "'");
            store.moveFile(path, destinationPath);
            if (destinationAlreadyExists)
               response.setStatus(HttpServletResponse.SC_NO_CONTENT);
            else
               response.setStatus(HttpServletResponse.SC_CREATED);
         }
         else if (store.isDirectory(path))
         {
            logger.debug("moving directory from '" + path + "' to '" + destinationPath + "'");
            store.moveDirectory(path, destinationPath);
            if (destinationAlreadyExists)
               response.setStatus(HttpServletResponse.SC_NO_CONTENT);
            else
               response.setStatus(HttpServletResponse.SC_CREATED);
         }
      }
      catch (CacheException e)
      {
         throw new ServletException(e);
      }
   }

   protected void doOptions(HttpServletRequest request, HttpServletResponse response) throws ServletException,
         IOException
   {
      response.addHeader("DAV", "1");
      response.addHeader("Allow", "COPY, DELETE, GET, HEAD, MKCOL, OPTIONS, PROPFIND, PROPPATCH, PUT");
      response.addHeader("MS-Author-Via", "DAV");
   }

   protected void doPut(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException
   {
      String path = getRelativePath(request);
      int fileSize = request.getContentLength();
      
      if (maxFileSize != -1 && fileSize > maxFileSize)
      {
         logger.warn("PUT: file "+ path +" ("+ fileSize +" bytes) exceeds max allowed size");
         response.sendError(HttpServletResponse.SC_FORBIDDEN, "Forbidden: max file size is " + maxFileSize + " bytes");
         return;
      }

      long freeMemory = Runtime.getRuntime().freeMemory();
      // reading and storing requires twice as much
      if (2 * fileSize > freeMemory)
      {
         logger.warn("PUT: not enough memory to store "+ path +" ("+ fileSize +" bytes)");
         response.sendError(SC_INSUFFICIENT_STORAGE, "Insufficient Storage");
         return;
      }

      try
      {
         logger.debug("creating file '" + path + "'");
         byte[] binaryContent = getBinaryContent(request.getInputStream());
         store.createFile(path, binaryContent);
         response.setStatus(HttpServletResponse.SC_CREATED);
      }
      catch (CacheException cacheException)
      {
         throw new ServletException(cacheException);
      }
      catch (OutOfMemoryError outOfMemoryError)
      {
         logger.error("PUT: OutOfMemoryError while storing "+ path +" ("+ fileSize +" bytes)", outOfMemoryError);
         response.sendError(SC_INSUFFICIENT_STORAGE, "Insufficient Storage");
         return;
      }
   }

   protected void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException
   {
      String relativePath = getRelativePath(request);
      if (!store.isFile(relativePath))
      {
         if (store.isDirectory(relativePath))
         {
            // TODO rfc2518 allows GET for Collections as well
            logger.warn("GET Collection not implemented: '" + relativePath + "'");
         }
         else
         {
            logger.debug("cannot get; file not found: '" + relativePath + "'");
         }
         response.setStatus(HttpServletResponse.SC_NOT_FOUND);
         return;
      }
      try
      {
         logger.debug("getting file '" + relativePath + "'");
         response.setContentType(store.getContentType(relativePath));
         byte[] content = store.getFileContent(relativePath);
         response.getOutputStream().write(content);
      }
      catch (CacheException e)
      {
         throw new ServletException(e);
      }
   }

   protected void doDelete(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException
   {
      String relativePath = getRelativePath(request);
      try
      {
         if (store.isDirectory(relativePath))
         {
            logger.debug("deleting directory '" + relativePath + "'");
            store.deleteDirectory(relativePath);
            response.setStatus(HttpServletResponse.SC_NO_CONTENT);
         }
         else if (store.isFile(relativePath))
         {
            logger.debug("deleting file '" + relativePath + "'");
            store.deleteFile(relativePath);
            response.setStatus(HttpServletResponse.SC_NO_CONTENT);
         }
         else
         {
            logger.debug("cannot delete; file not found: '" + relativePath + "'");
            response.setStatus(HttpServletResponse.SC_NOT_FOUND);
         }
      }
      catch (CacheException e)
      {
         throw new ServletException(e);
      }
   }

   protected void doMkCol(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException
   {
      String path = getRelativePath(request);
      if (store.exists(path))
      {
         logger.debug("cannot create directory; file already exists: '" + path + "'");
         response.sendError(HttpServletResponse.SC_METHOD_NOT_ALLOWED, "Method Not Allowed: collection already exists: " + path);
         return;
      }
      String parentPath = getParentPath(path);
      if (parentPath != null && !store.exists(parentPath))
      {
         logger.debug("cannot create directory; parent does not exist: '" + path + "'");
         response.sendError(HttpServletResponse.SC_CONFLICT, "Conflict: parent path does not exist");
         return;
      }
      if (request.getContentLength() > 0)
      {
         // we only support MKCOL with empty request body; see RFC2518:8.3.1
         response.sendError(HttpServletResponse.SC_UNSUPPORTED_MEDIA_TYPE, "Unsupported Media Type");
         return;
      }
      try
      {
         logger.debug("creating directory '" + path + "'");
         store.createDirectory(path);
         response.setStatus(HttpServletResponse.SC_CREATED);
      }
      catch (CacheException e)
      {
         throw new ServletException(e);
      }
   }

   protected void doPropFind(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException
   {
      int depth = getDepth(request);
      if (depth == DEPTH_INFINITY)
      {
         // can't see any good reason to implement this
         logger.warn("PROPFIND with 'Depth: infinity' not implemented");
         response.sendError(HttpServletResponse.SC_METHOD_NOT_ALLOWED, "Method Not Allowed: 'Depth: infinity' not supported");
         return;
      }

      String absolutePath = getAbsolutePath(request);
      String relativePath = getRelativePath(request);

      try
      {
         if (!store.exists(relativePath))
         {
            response.sendError(HttpServletResponse.SC_NOT_FOUND);
            return;
         }

         PropFindResponse xmlResponse = null;
         try
         {
            xmlResponse = new PropFindResponse(request);
         }
         catch (DocumentException e)
         {
            response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Bad Request: invalid XML");
            return;
         }
         xmlResponse.addResource(absolutePath, store.isDirectory(relativePath), store.getProperties(relativePath));
         if (depth == 1)
         {
            Set leaves = store.getFileNames(relativePath);
            if (leaves != null)
            {
               for (Iterator it = leaves.iterator(); it.hasNext();)
               {
                  String leafPath = (String) it.next();
                  xmlResponse.addResource(absolutePath + "/" + leafPath, false, store.getProperties(relativePath + "/"
                        + leafPath));
               }
            }
            Set branches = store.getDirectoryNames(relativePath);
            if (branches != null)
            {
               for (Iterator it = branches.iterator(); it.hasNext();)
               {
                  String branchPath = (String) it.next();
                  xmlResponse.addResource(absolutePath + "/" + branchPath, true, store.getProperties(relativePath + "/"
                        + branchPath));
               }
            }
         }
         xmlResponse.send(response);
      }
      catch (CacheException e)
      {
         throw new ServletException(e);
      }
   }

   protected void doPropPatch(HttpServletRequest request, HttpServletResponse response) throws ServletException,
         IOException
   {
      String absolutePath = getAbsolutePath(request);
      try
      {
         PropPatchResponse propPatchResponse = new PropPatchResponse(request);
         propPatchResponse.setResource(absolutePath);
         propPatchResponse.send(response);
      }
      catch (DocumentException e)
      {
         response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Bad Request: invalid XML");
         return;
      }
   }

   /**
    * E.g. "/myWebApp/a/b/c" returns "/a/b/c"; "/myWebApp" returns ""
    * 
    * @param request
    * @return
    */
   private String getRelativePath(HttpServletRequest request)
   {
      return relativizePath(request, getAbsolutePath(request));
   }

   private String relativizePath(HttpServletRequest request, String path)
   {
      int schema = path.indexOf("://");
      if (schema != -1)
      {
         // strip http://www.host.com/
         path = path.substring(path.indexOf("/", schema + 3));
      }
      String contextPath = request.getContextPath();
      String servletPath = request.getServletPath();
      return stripSlash(path.substring(contextPath.length() + servletPath.length()));
   }

   private String getAbsolutePath(HttpServletRequest request)
   {
      return stripSlash(request.getRequestURI());
   }

   private String getParentPath(String path)
   {
      int slash = path.lastIndexOf('/');
      if (slash == -1)
         return null;
      return path.substring(0, slash);
   }

   private String stripSlash(String path)
   {
      if (path.endsWith("/"))
      {
         return path.substring(0, path.length() - 1);
      }
      else
      {
         return path;
      }
   }

   private int getDepth(HttpServletRequest request)
   {
      int depth = DEPTH_INFINITY;
      String depthHeader = request.getHeader("Depth");
      if (depthHeader != null)
      {
         if (depthHeader.equals("0"))
         {
            depth = 0;
         }
         else if (depthHeader.equals("1"))
         {
            depth = 1;
         }
         else if (depthHeader.equals("infinity"))
         {
            depth = DEPTH_INFINITY;
         }
         else
         {
            logger.warn("unknown Depth: " + depthHeader);
         }
      }
      return depth;
   }

   private byte[] getBinaryContent(InputStream input) throws IOException
   {
      final int BUFFER_SIZE = 4 * 1024;
      byte[] buffer = new byte[BUFFER_SIZE];
      ByteArrayOutputStream output = new ByteArrayOutputStream();
      int count;
      while ((count = input.read(buffer)) != -1)
      {
         output.write(buffer, 0, count);
      }
      output.close();
      input.close();
      return output.toByteArray();
   }
}
