//
// Copyright (C) 2005 Mirko Nasato <mirko@artofsolving.com>
//
// Distributable under the terms of the LGPL license.
// See http://www.gnu.org/copyleft/lesser.html
//
package net.sf.davcache.functest;

import java.io.IOException;
import java.util.Enumeration;
import java.util.NoSuchElementException;

import junit.framework.TestCase;

import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpURL;
import org.apache.webdav.lib.ResponseEntity;
import org.apache.webdav.lib.WebdavResource;

public class WebdavServletTest extends TestCase
{
   private static final String BASE_PATH = "/davcache/dav";

   private static final String BASE_URL = "http://localhost:8080" + BASE_PATH;

   private WebdavResource root;

   protected void setUp() throws Exception
   {
      root = getWebdavResource("/");
   }

   protected void tearDown() throws Exception
   {
      root.deleteMethod(); // clean up everything
      root.close();
   }

   private static WebdavResource getWebdavResource(String relativePath) throws IOException
   {
      return new WebdavResource(new HttpURL(BASE_URL + relativePath));
   }

   private WebdavResource findResource(WebdavResource[] resources, String name)
   {
      for (int i = 0; i < resources.length; i++) {
         if (resources[i].getName().equals(name))
         {
            return resources[i];
         }
      }
      throw new NoSuchElementException("resource not found: " + name);
   }

   public void testCreateRenameAndDeleteDirectory() throws Exception
   {
      root.mkcolMethod(BASE_PATH + "/test");
      WebdavResource[] resources = root.listWebdavResources();
      assertEquals(1, resources.length);
      WebdavResource directory = resources[0];
      assertTrue(directory.isCollection());
      assertEquals("test", directory.getName());

      directory.moveMethod(BASE_PATH + "/test2");
      resources = root.listWebdavResources();
      assertEquals(1, resources.length);
      directory = resources[0];
      assertTrue(directory.isCollection());
      assertEquals("test2", directory.getName());

      directory.deleteMethod();
      resources = root.listWebdavResources();
      assertEquals(0, resources.length);
   }

   public void testCreateMoveAndDeleteFile() throws Exception
   {
      String testContent = "Test content.";
      root.putMethod(BASE_PATH + "/test.txt", testContent);
      WebdavResource[] resources = root.listWebdavResources();
      assertEquals(1, resources.length);
      WebdavResource file = resources[0];
      assertFalse(file.isCollection());
      assertEquals("test.txt", file.getName());
      assertEquals(testContent.length(), file.getGetContentLength());
      assertEquals("text/plain", file.getGetContentType());
      //TODO test LastModified
      assertEquals(testContent, file.getMethodDataAsString());

      // move inside a new folder
      root.mkcolMethod(BASE_PATH + "/folder");
      file.moveMethod(BASE_PATH + "/folder/moved.txt");
      resources = root.listWebdavResources();
      assertEquals(1, resources.length);
      WebdavResource folder = resources[0];
      assertTrue(folder.isCollection());
      assertEquals("folder", folder.getName());
      resources = folder.listWebdavResources();
      assertEquals(1, resources.length);
      file = resources[0];
      assertFalse(file.isCollection());
      assertEquals("moved.txt", file.getName());
      assertEquals(testContent, file.getMethodDataAsString());

      file.deleteMethod();
      resources = folder.listWebdavResources();
      assertEquals(0, resources.length);
   }

   public void testCopyFile() throws Exception
   {
      String testContent = "Test content.";
      root.putMethod(BASE_PATH + "/test.txt", testContent);
      WebdavResource[] resources = root.listWebdavResources();
      assertEquals(1, resources.length);
      WebdavResource file = resources[0];
      
      file.copyMethod(BASE_PATH + "/test2.txt");
      resources = root.listWebdavResources();
      assertEquals(2, resources.length);
      WebdavResource file1 = resources[1];
      assertEquals("test.txt", file1.getName());
      WebdavResource file2 = resources[0];
      assertEquals("test2.txt", file2.getName());
      assertEquals(testContent.length(), file.getGetContentLength());
      assertEquals("text/plain", file.getGetContentType());
      //TODO test LastModified
      assertEquals(testContent, file2.getMethodDataAsString());

      file1.deleteMethod();
      file2.deleteMethod();
      resources = root.listWebdavResources();
      assertEquals(0, resources.length);
   }

   public void testCopyDirectoryRecursively() throws Exception
   {
      root.mkcolMethod(BASE_PATH + "/folder1");
      root.putMethod(BASE_PATH + "/folder1/file1.txt", "file one");
      root.mkcolMethod(BASE_PATH + "/folder1/subfolder");
      root.putMethod(BASE_PATH + "/folder1/subfolder/file2.txt", "file two");
      
      root.setOverwrite(true);
      root.copyMethod(BASE_PATH + "/folder1", BASE_PATH + "/folder2");
      
      WebdavResource[] rootChildren = root.listWebdavResources();
      assertEquals("root should contain 2 children", 2, rootChildren.length);

      WebdavResource folder2 = findResource(rootChildren, "folder2");
      assertTrue("folder2 not a directory", folder2.exists() && folder2.isCollection());
      WebdavResource[] folder2Children = folder2.listWebdavResources();
      assertEquals("folder2 should contain 2 children", 2, folder2Children.length);

      WebdavResource file1 = findResource(folder2Children, "file1.txt");
      assertTrue("file1 not copied", file1.exists() && !file1.isCollection());
      assertEquals("file1 content not copied", "file one", file1.getMethodDataAsString());

      WebdavResource subfolder = findResource(folder2Children, "subfolder");
      assertTrue("subfolder not a directory", subfolder.exists() && subfolder.isCollection());
      WebdavResource[] subfolderChildren = subfolder.listWebdavResources();
      assertEquals("folder2 should contain 1 child", 1, subfolderChildren.length);

      WebdavResource file2 = subfolderChildren[0];
      assertTrue("file2 not copied", file2.exists() && !file2.isCollection());
      assertEquals("file2 content not copied", "file two", file2.getMethodDataAsString());
   }

   public void testMoveNonEmptyDirectory() throws Exception
   {
      root.mkcolMethod(BASE_PATH + "/folder1");
      root.putMethod(BASE_PATH + "/folder1/file1.txt", "file one");
      root.mkcolMethod(BASE_PATH + "/folder1/subfolder");
      root.putMethod(BASE_PATH + "/folder1/subfolder/file2.txt", "file two");
      
      root.setOverwrite(true);
      root.moveMethod(BASE_PATH + "/folder1", BASE_PATH + "/folder2");
      
      WebdavResource[] rootChildren = root.listWebdavResources();
      assertEquals("root should contain 1 child", 1, rootChildren.length);
      
      try {
         getWebdavResource("/folder1");
         fail("folder1 not deleted");
      } catch (HttpException notFound) {
         // expected
      }

      WebdavResource folder2 = findResource(rootChildren, "folder2");
      assertTrue("folder2 not a directory", folder2.exists() && folder2.isCollection());
      WebdavResource[] folder2Children = folder2.listWebdavResources();
      assertEquals("folder2 should contain 2 children", 2, folder2Children.length);

      WebdavResource file1 = findResource(folder2Children, "file1.txt");
      assertTrue("file1 not copied", file1.exists() && !file1.isCollection());
      assertEquals("file1 content not copied", "file one", file1.getMethodDataAsString());

      WebdavResource subfolder = findResource(folder2Children, "subfolder");
      assertTrue("subfolder not a directory", subfolder.exists() && subfolder.isCollection());
      WebdavResource[] subfolderChildren = subfolder.listWebdavResources();
      assertEquals("folder2 should contain 1 child", 1, subfolderChildren.length);

      WebdavResource file2 = subfolderChildren[0];
      assertTrue("file2 not copied", file2.exists() && !file2.isCollection());
      assertEquals("file2 content not copied", "file two", file2.getMethodDataAsString());
   }

   public void testPropFindAll() throws IOException
   {
      Enumeration entities = root.propfindMethod(1);
      assertTrue("no resources", entities.hasMoreElements());
      ResponseEntity entity = (ResponseEntity) entities.nextElement();
      assertEquals("request not successful", 200, entity.getStatusCode());
      Enumeration properties = entity.getProperties();
      assertTrue("no properties", properties.hasMoreElements());
   }
}
